{% extends "!layout.html" %}

{% block extrahead %}
{{ super() }}
<link rel="preconnect" href="https://fonts.googleapis.com">
<link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
<link href="https://fonts.googleapis.com/css2?family=Roboto+Mono&display=swap" rel="stylesheet">
<script async src="https://www.googletagmanager.com/gtag/js?id=G-YN4QSRPV0K"></script>
<script>
  window.dataLayer = window.dataLayer || [];
  function gtag(){dataLayer.push(arguments);}
  gtag('js', new Date());

  gtag('config', 'G-YN4QSRPV0K');
</script>
<script>
  (function() {
    if (typeof window === 'undefined') return;
    if (typeof window.signals !== 'undefined') return;
    var script = document.createElement('script');
    script.src = 'https://cdn.cr-relay.com/v1/site/9bf5cfb7-d3ce-4f5c-9de2-0802bea74edb/signals.js';
    script.async = true;
    window.signals = Object.assign(
      [],
      ['page', 'identify', 'form'].reduce(function (acc, method){
        acc[method] = function () {
          signals.push([method, arguments]);
          return signals;
        };
       return acc;
      }, {})
    );
    document.head.appendChild(script);
  })();
</script>
{% endblock %}

{% block docs_main %}
{{ super() }}

<!-- Mobile menu that pops up when hamburger menu is clicked -->
{% include "sections/mobile-menu.html" %}
<script src="{{ pathto('_static/mobile-menu.js', 1) }}"></script>
<script type="text/javascript">
    mobileMenu.bind();
</script>
{% endblock %}
