<div class="header-container">
    <img referrerpolicy="no-referrer-when-downgrade" src="https://static.scarf.sh/a.png?x-pxid=6f9dd415-f331-4826-97a7-83e94dc0f89e" />
    <a href="/"><img class="header-logo" src="{{ pathto('_static/daft-logo.png', 1) }}" alt="Daft logo" /></a>
    <nav class="header-nav">
        <ul class="header-nav-list">
            <li class="header-nav-listitem">
                <a href="https://www.getdaft.io/projects/docs/en/stable/quickstart/index.html">
                    Get Started
                </a>
            </li>
            <li class="header-nav-listitem">
                <a href="https://www.getdaft.io/projects/docs/en/stable/index.html">
                    User Guide
                </a>
            </li>
            <li class="header-nav-listitem">
                <a href="https://www.github.com/Eventual-Inc/Daft">
                    Github
                </a>
            </li>
            <li class="header-nav-listitem">
                <a href="https://blog.getdaft.io/">
                    Blog
                </a>
            </li>
            <li class="header-nav-listitem">
                <a href="https://www.getdaft.io/#community">
                    Community
                </a>
            </li>
        </ul>
    </nav>
    <a class="main-menu-open-button" href="#" data-behavior="open-mobile-menu"></a>
</div>
