<div class="mobile-main-menu">
    <div class="header-container">
        <a href="/"><img class="header-logo" src="{{ pathto('_static/daft-logo.png', 1) }}" alt="Daft logo" /></a>
        <a href="#" class="main-menu-close-button" data-behavior="close-mobile-menu"></a>
    </div>
    <ul class="header-nav-list mobile">
        <li class="header-nav-listitem mobile">
            <a href="https://www.getdaft.io/projects/docs/en/stable/quickstart/index.html">
                Get Started
            </a>
        </li>
        <li class="header-nav-listitem mobile">
            <a href="https://www.getdaft.io/projects/docs/en/stable/index.html">
                User Guide
            </a>
        </li>
        <li class="header-nav-listitem mobile">
            <a href="https://www.github.com/Eventual-Inc/Daft">
                Github
            </a>
        </li>
        <li class="header-nav-listitem mobile">
            <a href="https://blog.getdaft.io/">
                Blog
            </a>
        </li>
        <li class="header-nav-listitem mobile">
            <a href="https://www.getdaft.io/#community">
                Community
            </a>
        </li>
    </ul>
</div>
