/* Taken from suggested solutions in: https://github.com/sphinx-doc/sphinx/issues/1514 */

/* Newlines (\a) and spaces (\20) before each parameter */
.sig-param::before {
    content: "\a\20\20\20\20";
    white-space: pre;
}

/* Newline after the last parameter (so the closing bracket is on a new line) */
dt em.sig-param:last-of-type::after {
    content: "\a";
    white-space: pre;
}
