from daft.daft import ImageFormat, ImageMode, PySeries

def decode(s: PySeries, raise_error_on_failure: bool, mode: ImageMode | None = None) -> PySeries: ...
def encode(s: PySeries, image_format: ImageFormat) -> PySeries: ...
def resize(s: PySeries, w: int, h: int) -> PySeries: ...
def to_mode(s: PySeries, mode: ImageMode) -> PySeries: ...
