.header-container {
    display: flex;
    flex-direction: row;
    align-items: center;
    justify-content: space-between;
    padding: 24px;
    border-bottom: 1px solid rgba(0,0,0,.1);
    width: 100%;
}

.header-nav {
    flex-grow: 1;
    /* Set to block when media width large */
    display: none;
}

.header-nav-list {
    display: flex;
    flex-direction: row;
    margin: 0;
    justify-content: space-around;
}
.header-nav-list.mobile {
    padding: 24px;
    flex-direction: column;
    gap: 8px;
}
.header-nav-listitem {
    list-style: none;
    font-weight: 450;
}
.header-nav-listitem.mobile {
    font-size: 24px;
}

.header-logo {
    height: 36px;
    width: auto;
}


/* Hamburger menu displayed when screen width is small */
.main-menu-open-button {
    background-image: url(icon-menu-dots.svg);
    background-position: center center;
    background-size: 25px 25px;
    background-repeat: no-repeat;
    width: 25px;
    height: 25px;
}
.main-menu-close-button {
    background-image: url(icon-menu-close.svg);
    background-position: center center;
    background-size: 25px 25px;
    background-repeat: no-repeat;
    width: 25px;
    height: 25px;
}

/* Fullscreen mobile menu displayed when main-menu-open-button is clicked */
.mobile-main-menu {
    display: none;
}
.mobile-main-menu.open {
    display: flex;
    flex-direction: column;
    background-color: #fcfcfc;
    position: fixed;
    z-index: 9999;
    left: 0;
    right: 0;
    top: 0;
    min-height: 100%;
    height: 100%;
    width: 100%;
    margin-left: auto;
    margin-right: auto;
}

/* Applied when mobile menu button is clicked */
body.no-scroll {
    height: 100%;
    overflow: hidden;
}

@media(min-width: 768px) {
    .main-menu-open-button {
        display: none;
    }
    .header-nav {
        display: block;
    }
    .header-container {
        padding-left: 10vw;
        padding-right: 10vw;
        width: 100%;
    }
}
